@@ -11,19 +11,22 @@ import binascii
 import shutil
 import tempfile
 import unittest
+from array import array
 from string import hexdigits
 from random import choice, randint, random
 from collections import Counter
 
-from bitarray import (bitarray, frozenbitarray, decodetree,
+from bitarray import (bitarray, frozenbitarray, decodetree, bits2bytes,
                       get_default_endian, _set_default_endian)
 from bitarray.test_bitarray import Util, skipIf
 
 from bitarray.util import (
     zeros, urandom, pprint, make_endian, rindex, strip, count_n,
-    parity, count_and, count_or, count_xor, subset,
+    parity, count_and, count_or, count_xor, any_and, subset, _correspond_all,
+    intervals,
     serialize, deserialize, ba2hex, hex2ba, ba2base, base2ba,
-    ba2int, int2ba, vl_encode, vl_decode,
+    ba2int, int2ba,
+    sc_encode, sc_decode, vl_encode, vl_decode,
     huffman_code, canonical_huffman, canonical_decode,
 )
 
@@ -41,13 +44,10 @@ class TestsZeros(unittest.TestCase):
     def test_basic(self):
         for default_endian in 'big', 'little':
             _set_default_endian(default_endian)
-            a = zeros(0)
-            self.assertEqual(a, bitarray())
-            self.assertEqual(a.endian(), default_endian)
 
-            a = zeros(0, endian=None)
-            self.assertEqual(len(a), 0)
-            self.assertEqual(a.endian(), default_endian)
+            for a in zeros(0), zeros(0, None), zeros(0, endian=None):
+                self.assertEqual(a, bitarray())
+                self.assertEqual(a.endian(), default_endian)
 
             for n in range(100):
                 a = zeros(n)
@@ -55,11 +55,12 @@ class TestsZeros(unittest.TestCase):
                 self.assertFalse(a.any())
                 self.assertEqual(a.count(1), 0)
                 self.assertEqual(a, bitarray(n * '0'))
+                self.assertIsInstance(a, bitarray)
 
             for endian in 'big', 'little':
-                a = zeros(3, endian)
-                self.assertEqual(a, bitarray('000'))
-                self.assertEqual(a.endian(), endian)
+                for a in zeros(3, endian), zeros(3, endian=endian):
+                    self.assertEqual(a, bitarray('000'))
+                    self.assertEqual(a.endian(), endian)
 
     def test_wrong_args(self):
         self.assertRaises(TypeError, zeros) # no argument
@@ -84,13 +85,10 @@ class TestsURandom(unittest.TestCase):
     def test_basic(self):
         for default_endian in 'big', 'little':
             _set_default_endian(default_endian)
-            a = urandom(0)
-            self.assertEqual(a, bitarray())
-            self.assertEqual(a.endian(), default_endian)
 
-            a = urandom(7, endian=None)
-            self.assertEqual(len(a), 7)
-            self.assertEqual(a.endian(), default_endian)
+            for a in urandom(0), urandom(0, None), urandom(0, endian=None):
+                self.assertEqual(a, bitarray())
+                self.assertEqual(a.endian(), default_endian)
 
             for n in range(50):
                 a = urandom(n)
@@ -98,9 +96,9 @@ class TestsURandom(unittest.TestCase):
                 self.assertEqual(a.endian(), default_endian)
 
             for endian in 'big', 'little':
-                a = urandom(11, endian)
-                self.assertEqual(len(a), 11)
-                self.assertEqual(a.endian(), endian)
+                for a in urandom(11, endian), urandom(11, endian=endian):
+                    self.assertEqual(len(a), 11)
+                    self.assertEqual(a.endian(), endian)
 
     def test_count(self):
         a = urandom(1000)
@@ -609,10 +607,86 @@ class TestsBitwiseCount(unittest.TestCase, Util):
             self.assertEqual(count_or(a, b),  (a | b).count())
             self.assertEqual(count_xor(a, b), (a ^ b).count())
 
+    def test_misc(self):
+        for a in self.randombitarrays():
+            n = len(a)
+            b = urandom(n, a.endian())
+            # any and
+            self.assertEqual(any(a & b), count_and(a, b) > 0)
+            self.assertEqual(any_and(a, b), any(a & b))
+            # any or
+            self.assertEqual(any(a | b), count_or(a, b) > 0)
+            self.assertEqual(any(a | b), any(a) or any(b))
+            # any xor
+            self.assertEqual(any(a ^ b), count_xor(a, b) > 0)
+            self.assertEqual(any(a ^ b), a != b)
+
+            # all and
+            self.assertEqual(all(a & b), count_and(a, b) == n)
+            self.assertEqual(all(a & b), all(a) and all(b))
+            # all or
+            self.assertEqual(all(a | b), count_or(a, b) == n)
+            # all xor
+            self.assertEqual(all(a ^ b), count_xor(a, b) == n)
+            self.assertEqual(all(a ^ b), a == ~b)
+
 tests.append(TestsBitwiseCount)
 
 # ---------------------------------------------------------------------------
 
+class TestsBitwiseAny(unittest.TestCase, Util):
+
+    def test_basic(self):
+        a = frozenbitarray('0101')
+        b = bitarray('0111')
+        self.assertTrue(any_and(a, b))
+        self.assertRaises(TypeError, any_and)
+        self.assertRaises(TypeError, any_and, a, 4)
+        b.append(1)
+        self.assertRaises(ValueError, any_and, a, b)
+        self.assertRaises(ValueError, any_and,
+                          bitarray('01', 'little'),
+                          bitarray('11', 'big'))
+
+    def check(self, a, b):
+        r = any_and(a, b)
+        self.assertIsInstance(r, bool)
+        self.assertEqual(r, any_and(b, a))  # symmetry
+        self.assertEqual(r, any(a & b))
+        self.assertEqual(r, (a & b).any())
+        self.assertEqual(r, count_and(a, b) > 0)
+
+    def test_explitcit(self):
+        for a, b , res in [
+                ('', '', False),
+                ('0', '1', False),
+                ('0', '0', False),
+                ('1', '1', True),
+                ('00011', '11100', False),
+                ('00001011 1', '01000100 1', True)]:
+            a = bitarray(a)
+            b = bitarray(b)
+            self.assertTrue(any_and(a, b) is res)
+            self.check(a, b)
+
+    def test_random(self):
+        for a in self.randombitarrays():
+            n = len(a)
+            b = urandom(n, a.endian())
+            self.check(a, b)
+
+    def test_one(self):
+        for n in range(1, 200):
+            a = zeros(n)
+            b = urandom(n)
+            i = randint(0, n - 1)
+            a[i] = 1
+            self.assertEqual(b[i], any_and(a, b))
+
+tests.append(TestsBitwiseAny)
+
+# ---------------------------------------------------------------------------
+
 class TestsSubset(unittest.TestCase, Util):
 
     def test_basic(self):
@@ -664,6 +738,45 @@ tests.append(TestsSubset)
 
 # ---------------------------------------------------------------------------
 
+class TestsCorrespondAll(unittest.TestCase, Util):
+
+    def test_basic(self):
+        a = frozenbitarray('0101')
+        b = bitarray('0111')
+        self.assertTrue(_correspond_all(a, b), (1, 1, 1, 1))
+        self.assertRaises(TypeError, _correspond_all)
+        b.append(1)
+        self.assertRaises(ValueError, _correspond_all, a, b)
+        self.assertRaises(ValueError, _correspond_all,
+                          bitarray('01', 'little'),
+                          bitarray('11', 'big'))
+
+    def test_explitcit(self):
+        for a, b, res in [
+                ('', '', (0, 0, 0, 0)),
+                ('0000011111',
+                 '0000100111', (4, 1, 2, 3)),
+            ]:
+            self.assertEqual(_correspond_all(bitarray(a), bitarray(b)), res)
+
+    def test_random(self):
+        for a in self.randombitarrays():
+            n = len(a)
+            b = urandom(n, a.endian())
+            res = _correspond_all(a, b)
+            self.assertEqual(res[0], count_and(~a, ~b))
+            self.assertEqual(res[1], count_and(~a, b))
+            self.assertEqual(res[2], count_and(a, ~b))
+            self.assertEqual(res[3], count_and(a, b))
+
+            self.assertEqual(res[0], n - count_or(a, b))
+            self.assertEqual(res[1] + res[2], count_xor(a, b))
+            self.assertEqual(sum(res), n)
+
+tests.append(TestsCorrespondAll)
+
+# ---------------------------------------------------------------------------
+
 class TestsParity(unittest.TestCase, Util):
 
     def test_bitarray(self):
@@ -703,6 +816,68 @@ tests.append(TestsParity)
 
 # ---------------------------------------------------------------------------
 
+class TestsIntervals(unittest.TestCase, Util):
+
+    def test_explicit(self):
+        for s, lst in [
+                ('', []),
+                ('0', [(0, 0, 1)]),
+                ('1', [(1, 0, 1)]),
+                ('00111100 00000111 00',
+                 [(0, 0, 2), (1, 2, 6), (0, 6, 13), (1, 13, 16), (0, 16, 18)]),
+            ]:
+            a = bitarray(s)
+            self.assertEqual(list(intervals(a)), lst)
+
+    def test_count(self):
+        for s, res in [
+                ('', 0),
+                ('0', 1),
+                ('1', 1),
+                ('00', 1),
+                ('01', 2),
+                ('10', 2),
+                ('11', 1),
+                ('0011110000000', 3),
+            ]:
+            a = bitarray(s)
+            self.assertEqual(res, len(list(intervals(a))))
+            self.assertEqual(res, sum(1 for _ in intervals(a)))
+
+    def test_random(self):
+        for a in self.randombitarrays():
+            b = urandom(len(a))
+            cnt = [0, 0]
+            v = a[0] if a else None
+            for value, start, stop in intervals(a):
+                self.assertFalse(isinstance(value, bool))
+                self.assertEqual(value, v)
+                v = not v
+                self.assertTrue(0 <= start < stop <= len(a))
+                cnt[value] += stop - start
+                b[start:stop] = value
+            self.assertEqual(a, b)
+            for v in 0, 1:
+                self.assertEqual(cnt[v], a.count(v))
+
+    def test_runs(self):
+        for a in self.randombitarrays():
+            first = a[0] if a else None
+            # list runs of alternating bits
+            runs = [stop - start for _, start, stop in intervals(a)]
+
+            b = bitarray()
+            v = first
+            for length in runs:
+                b.extend(length * bitarray([v]))
+                v = not v
+
+            self.assertEqual(a, b)
+
+tests.append(TestsIntervals)
+
+# ---------------------------------------------------------------------------
+
 class TestsHexlify(unittest.TestCase, Util):
 
     def test_ba2hex(self):
@@ -735,6 +910,7 @@ class TestsHexlify(unittest.TestCase, Util):
             a = hex2ba(c)
             self.assertEqual(a.to01(), '1110')
             self.assertEqual(a.endian(), 'big')
+            self.assertIsType(a, 'bitarray')
         self.assertEQUAL(hex2ba('01'), bitarray('0000 0001', 'big'))
         self.assertEQUAL(hex2ba('08', 'little'),
                          bitarray('0000 0001', 'little'))
@@ -809,6 +985,7 @@ class TestsBase(unittest.TestCase, Util):
             a = base2ba(16, c)
             self.assertEqual(a.to01(), '1110')
             self.assertEqual(a.endian(), 'big')
+            self.assertIsType(a, 'bitarray')
 
     def test_explicit(self):
         data = [ #              n  little   big
@@ -857,6 +1034,9 @@ class TestsBase(unittest.TestCase, Util):
         self.assertRaises(TypeError, ba2base, 32, None)
         self.assertRaises(TypeError, base2ba, 32, None)
 
+        for i in 2, 4, 8, 16, 32, 64:
+            self.assertRaises(UnicodeEncodeError, base2ba, i, u'10\u20ac')
+
     def test_binary(self):
         a = base2ba(2, '1011')
         self.assertEqual(a, bitarray('1011'))
@@ -953,6 +1133,275 @@ tests.append(TestsBase)
 
 # ---------------------------------------------------------------------------
 
+class SC_Tests(unittest.TestCase, Util):
+
+    def test_explicit(self):
+        for b, bits, endian in [
+                (b'\x00\0',                 '',                  'little'),
+                (b'\x01\x03\x01\x03\0',     '110',               'little'),
+                (b'\x11\x07\x01\x02\0',     '0000001',           'big'),
+                (b'\x01\x10\x02\xf0\x0f\0', '00001111 11110000', 'little'),
+                (b'\x11\x10\xa1\x0c\0',     '00000000 00001000', 'big'),
+                (b'\x11\x09\xa1\x08\0',     '00000000 1',        'big'),
+                (b'\x01E\xa3ABD\0',         65 * '0' + '1101',   'little'),
+        ]:
+            # the padbits in a frozenbitarray are guaranteed to be zero
+            a = frozenbitarray(bits, endian)
+            self.assertEqual(sc_encode(a), b)
+            self.assertEqual(sc_decode(b), a)
+
+    def test_decode_header_nbits(self):
+        for b, n in [
+                (b'\x00\0', 0),
+                (b'\x01\x00\0', 0),
+                (b'\x01\x01\0', 1),
+                (b'\x02\x00\x00\0', 0),
+                (b'\x02\x00\x01\0', 256),
+                (b'\x03\x00\x00\x00\0', 0),
+                (b'\x03\x00\x00\x01\0', 65536),
+        ]:
+            a = sc_decode(b)
+            self.assertEqual(len(a), n)
+            self.assertFalse(a.any())
+
+    @skipIf(sys.version_info[0] == 2)
+    def test_decode_untouch(self):
+        stream = iter(b'\x01\x03\x01\x03\0XYZ')
+        self.assertEqual(sc_decode(stream), bitarray('110'))
+        self.assertEqual(next(stream), ord('X'))
+
+        stream = iter([0x11, 0x05, 0x01, 0xff, 0, None, 'foo'])
+        self.assertEqual(sc_decode(stream), bitarray('11111'))
+        self.assertTrue(next(stream) is None)
+        self.assertEqual(next(stream), 'foo')
+
+    @skipIf(sys.version_info[0] == 2)
+    def test_decode_header_errors(self):
+        # invalid header
+        for c in 0x20, 0x21, 0x40, 0x80, 0xc0, 0xf0, 0xff:
+            self.assertRaisesMessage(ValueError,
+                                     "invalid header: 0x%02x" % c,
+                                     sc_decode,
+                                     bytearray([c]))
+        # invalid block head
+        for c in 0x81, 0x9f, 0xc0, 0xc1, 0xc5, 0xff:
+            self.assertRaisesMessage(ValueError,
+                                     "invalid block head: 0x%02x" % c,
+                                     sc_decode,
+                                     bytearray([0x01, 0x10, c]))
+
+    def test_decode_header_overflow(self):
+        nbytes = tuple.__itemsize__
+        self.assertRaisesMessage(
+            OverflowError,
+            "sizeof(Py_ssize_t) = %d: cannot read 9 bytes" % nbytes,
+            sc_decode, b'\x09' + 9 * b'\x00')
+
+        self.assertRaisesMessage(
+            ValueError,
+            "read %d bytes got negative value: -1" % nbytes,
+            sc_decode, bytes(bytearray([nbytes] + nbytes * [0xff])))
+
+        if nbytes == 4:
+            self.assertRaisesMessage(
+                OverflowError,
+                "sizeof(Py_ssize_t) = 4: cannot read 5 bytes",
+                sc_decode, b'\x05' + 5 * b'\x00')
+
+            self.assertRaisesMessage(
+                ValueError,
+                "read 4 bytes got negative value: -2147483648",
+                sc_decode, b'\x04\x00\x00\x00\x80')
+
+    def test_decode_errors(self):
+        # too many raw bytes
+        self.assertRaisesMessage(
+            ValueError, "decode error (raw): 0 + 2 > 1",
+            sc_decode, b"\x01\x05\x02\xff\xff\0")
+        self.assertRaisesMessage(
+            ValueError, "decode error (raw): 32 + 3 > 34",
+            sc_decode, b"\x02\x0f\x01\xa0\x03\xff\xff\xff\0")
+        # sparse index too high
+        self.assertRaisesMessage(
+            ValueError, "decode error (n=1): 128 >= 128",
+            sc_decode, b"\x01\x80\xa1\x80\0")
+        self.assertRaisesMessage(
+            ValueError, "decode error (n=2): 512 >= 512",
+            sc_decode, b"\x02\x00\x02\xc2\x01\x00\x02\0")
+        self.assertRaisesMessage(
+            ValueError, "decode error (n=3): 32768 >= 32768",
+            sc_decode, b"\x02\x00\x80\xc3\x01\x00\x80\x00\0")
+
+        if tuple.__itemsize__ == 4:
+            msg = "read 4 bytes got negative value: -2147483648"
+        else:
+            msg = "decode error (n=4): 2147483648 >= 16"
+        self.assertRaisesMessage(
+            ValueError, msg,
+            sc_decode, b"\x01\x10\xc4\x01\x00\x00\x00\x80\0")
+
+        if tuple.__itemsize__ == 4:
+            msg = "read 4 bytes got negative value: -1"
+        else:
+            msg = "decode error (n=4): 4294967295 >= 16"
+        self.assertRaisesMessage(
+            ValueError, msg,
+            sc_decode, b"\x01\x10\xc4\x01\xff\xff\xff\xff\0")
+
+    def test_decode_end_of_stream(self):
+        for stream in [b'', b'\x00', b'\x01', b'\x02\x77',
+                       b'\x01\x04\x01', b'\x01\x04\xa1', b'\x01\x04\xa0']:
+            self.assertRaisesMessage(ValueError, "unexpected end of stream",
+                                     sc_decode, stream)
+
+    @skipIf(sys.version_info[0] == 2)
+    def test_decode_types(self):
+        blob = b'\x11\x03\x01\x20\0'
+        for b in blob, bytearray(blob), list(blob), array('B', blob):
+            a = sc_decode(b)
+            self.assertIsType(a, 'bitarray')
+            self.assertEqual(a.endian(), 'big')
+            self.assertEqual(a.to01(), '001')
+
+        self.assertRaises(TypeError, sc_decode, [0x02, None])
+        for x in None, 3, 3.2, Ellipsis:
+            self.assertRaises(TypeError, sc_decode, x)
+        for _ in range(10):
+            self.assertRaises(TypeError, sc_decode, [0x00, None])
+
+    def test_decode_ambiguity(self):
+        for b in [
+                # raw:
+                b'\x11\x03\x01\x20\0',    # this is what sc_encode gives us
+                b'\x11\x03\x01\x2f\0',    # some pad bits are 1
+                # sparse:
+                b'\x11\x03\xa1\x02\0',                  # using block type 1
+                b'\x11\x03\xc2\x01\x02\x00\0',          # using block type 2
+                b'\x11\x03\xc3\x01\x02\x00\x00\0',      # using block type 3
+                b'\x11\x03\xc4\x01\x02\x00\x00\x00\0',  # using block type 4
+        ]:
+            a = sc_decode(b)
+            self.assertEqual(a.to01(), '001')
+
+    @skipIf(sys.version_info[0] == 2)
+    def test_raw_block(self):
+        a = bitarray(0, 'big')
+        for nbytes in range(1, 129):
+            nbits = 8 * nbytes
+            raw = nbytes * b'\xf7'
+            if nbits < 256:
+                b = bytearray([0x11, nbits])
+            else:
+                b = bytearray([0x12, nbits & 0xff, nbits >> 8])
+            b.append(nbytes)  # head byte
+            b.extend(raw)
+            b.append(0)       # stop byte
+
+            a.frombytes(b'\xf7')
+            self.assertEqual(sc_decode(b), a)
+            self.assertEqual(sc_encode(a), bytes(b))
+
+    @skipIf(sys.version_info[0] == 2)
+    def test_sparse_block_type1(self):
+        a = bitarray(256, 'little')
+        for n in range(1, 32):
+            positions = os.urandom(n)
+            b = bytearray([0x02, 0x00, 0x01, 0xa0 + n])
+            b.extend(positions)
+            b.append(0)  # stop byte
+
+            a.setall(0)
+            for p in positions:
+                a[p] = 1
+            self.assertEqual(sc_decode(b), a)
+
+            # in order to recreate the block sc_encode generates, we need
+            # a sorted list of the positions with no duplicates
+            lst = sorted(set(positions))
+            b = bytearray([0x02, 0x00, 0x01, 0xa0 + len(lst)])
+            b.extend(lst)
+            b.append(0)  # stop
+
+            self.assertEqual(sc_decode(b), a)
+            self.assertEqual(sc_encode(a), bytes(b))
+
+    def test_decode_random_bytes(self):
+        # ensure random input doesn't crash the decoder
+        cnt = 0
+        for _ in range(5000):
+            n = randint(0, 20)
+            b = b'\x02\x00\x04' + os.urandom(n)
+            try:
+                a = sc_decode(b)
+            except ValueError as e:
+                if e != 'unexpected end of stream':
+                    continue
+            self.assertEqual(len(a), 1024)
+            self.assertEqual(a.endian(), 'little')
+            cnt += 1
+        self.assertTrue(cnt > 2)
+
+    def test_encode_types(self):
+        for a in bitarray('1', 'big'), frozenbitarray('1', 'big'):
+            b = sc_encode(a)
+            self.assertIsInstance(b, bytes)
+            self.assertEqual(len(b), 5)
+
+        for a in None, [], 0, 123, b'', b'\x00', 3.14:
+            self.assertRaises(TypeError, sc_encode, a)
+
+    def round_trip(self, a):
+        b = sc_encode(a)
+        c = sc_decode(b)
+        self.assertEqual(a, c)
+        self.assertEqual(a.endian(), c.endian())
+
+    def test_encode_zeros(self):
+        for i in range(18):
+            n = 1 << i
+            a = bitarray(n)
+            a.setall(0)
+            m = 2                            # head byte and stop byte
+            m += bits2bytes(n.bit_length())  # size bytes
+            # For smaller or equal to 1 << 32, we have only one type 4 block.
+            m += bool(n > 0)                 # number of blocks (head bytes)
+            m += bool(n > 256)               # number of second block heads
+            b = sc_encode(a)
+            self.assertEqual(m, len(b))
+            self.round_trip(a)
+
+        a = zeros(1 << 25)
+        a[0] = 1
+        self.assertEqual(
+            sc_encode(a),
+            b'\x14\x00\x00\x00\x02\xc4\x01\x00\x00\x00\x00\x00')
+
+    def test_encode_ones(self):
+        for _ in range(50):
+            n = randint(0, 10000)
+            a = bitarray(n)
+            a.setall(1)
+            m = 2                            # head byte and stop byte
+            m += bits2bytes(n.bit_length())  # size bytes
+            m += (n + 1023) // 1024          # number of blocks (head bytes)
+            m += bits2bytes(n)               # actual raw bytes
+            self.assertEqual(m, len(sc_encode(a)))
+            self.round_trip(a)
+
+    def test_random(self):
+        for _ in range(10):
+            n = randint(0, 100000)
+            endian = self.random_endian()
+            a = bitarray(n, endian)
+            a.setall(1)
+            for _ in range(16):
+                a &= urandom(n, endian)
+                self.round_trip(a)
+
+tests.append(SC_Tests)
+
+# ---------------------------------------------------------------------------
+
 class VLFTests(unittest.TestCase, Util):
 
     def test_explicit(self):
@@ -998,15 +1447,25 @@ class VLFTests(unittest.TestCase, Util):
             lst.append(bytearray(b))
         for s in lst:
             a = vl_decode(s, endian=self.random_endian())
-            self.assertIsInstance(a, bitarray)
+            self.assertIsType(a, 'bitarray')
             self.assertEqual(a, bitarray('0011 01'))
 
     def test_decode_endian(self):
-        for endian in 'little', 'big', None:
-            a = vl_decode(b'\xd3\x20', endian)
-            self.assertEqual(a, bitarray('0011 01'))
-            self.assertEqual(a.endian(),
-                             endian if endian else get_default_endian())
+        blob = b'\xd3\x20'
+        res = bitarray('0011 01')
+
+        for default_endian in 'little', 'big':
+            _set_default_endian(default_endian)
+
+            for endian in 'little', 'big', None:
+                a = vl_decode(blob, endian)
+                self.assertEqual(a, res)
+                self.assertEqual(a.endian(),
+                                 endian if endian else default_endian)
+
+            a = vl_decode(blob)
+            self.assertEqual(a, res)
+            self.assertEqual(a.endian(), default_endian)
 
     def test_decode_trailing(self):
         for s, bits in [(b'\x40ABC', ''),
@@ -1034,26 +1493,14 @@ class VLFTests(unittest.TestCase, Util):
 
     def test_decode_errors(self):
         # decode empty bits
-        self.assertRaises(StopIteration, vl_decode, b'')
+        self.assertRaises(ValueError, vl_decode, b'')
         # invalid number of padding bits
         for s in b'\x50', b'\x60', b'\x70':
             self.assertRaises(ValueError, vl_decode, s)
         self.assertRaises(ValueError, vl_decode, b'\xf0')
         # high bit set, but no terminating byte
         for s in b'\x80', b'\x80\x80':
-            self.assertRaises(StopIteration, vl_decode, s)
-
-    def test_decode_error_message(self):
-        pat = re.compile(r'[\w\s,]+:\s+(\d+)')
-        for n in range(120):
-            a = None
-            s = bytes(bytearray([randint(0x80, 0xef) for _ in range(n)]))
-            try:
-                a = vl_decode(s)
-            except StopIteration as e:
-                m = pat.match(str(e))
-                self.assertEqual(m.group(1), str(n))
-            self.assertTrue(a is None)
+            self.assertRaises(ValueError, vl_decode, s)
 
     @skipIf(sys.version_info[0] == 2)
     def test_decode_invalid_stream(self):
@@ -1405,27 +1852,24 @@ tests.append(MixedTests)
 class TestsSerialization(unittest.TestCase, Util):
 
     def test_explicit(self):
-        for a, b in [
-                (bitarray(0, 'little'),   b'\x00'),
-                (bitarray(0, 'big'),      b'\x10'),
-                (bitarray('1', 'little'), b'\x07\x01'),
-                (bitarray('1', 'big'),    b'\x17\x80'),
-                (bitarray('11110000', 'little'), b'\x00\x0f'),
-                (bitarray('11110000', 'big'),    b'\x10\xf0'),
+        for blob, endian, bits in [
+                (b'\x00',         'little', ''),
+                (b'\x07\x01',     'little', '1'),
+                (b'\x17\x80',     'big',    '1'),
+                (b'\x13\xf8',     'big',    '11111'),
+                (b'\x00\x0f',     'little', '11110000'),
+                (b'\x10\xf0',     'big',    '11110000'),
+                (b'\x12\x87\xd8', 'big',    '10000111 110110')
         ]:
-            self.assertEqual(serialize(a), b)
-            self.assertEQUAL(deserialize(b), a)
+            a = bitarray(bits, endian)
+            s = serialize(a)
+            self.assertEqual(blob, s)
+            self.assertIsInstance(s, bytes)
 
-    def test_zeros_and_ones(self):
-        for endian in 'little', 'big':
-            for n in range(50):
-                a = zeros(n, endian)
-                s = serialize(a)
-                self.assertIsInstance(s, bytes)
-                self.assertEqual(s[1:], b'\0' * a.nbytes)
-                self.assertEQUAL(a, deserialize(s))
-                a.setall(1)
-                self.assertEQUAL(a, deserialize(serialize(a)))
+            b = deserialize(blob)
+            self.assertEqual(b, a)
+            self.assertEqual(b.endian(), endian)
+            self.assertIsType(b, 'bitarray')
 
     def test_serialize_args(self):
         for x in '0', 0, 1, b'\x00', 0.0, [0, 1], bytearray([0]):
@@ -1435,23 +1879,24 @@ class TestsSerialization(unittest.TestCase, Util):
         # too many arguments
         self.assertRaises(TypeError, serialize, bitarray(), 1)
 
-    # On Python 2, bytes(x) is the string representation of x,
-    # so x can almost be of any type.  And the string representation
-    # of these object will cause ValueErrors in bitarray(x).
-    # Instead of adding special cases in deserialize() or here,
-    # we decided to skip these tests for Python 2 altogether.
-    @skipIf(sys.version_info[0] == 2)
-    def test_deserialize_args(self):
-        for x in 0, 1, False, True, None, u'', u'01', 0.0, [0, 0.6]:
-            self.assertRaises(TypeError, deserialize, x)
-        self.assertRaises(TypeError, deserialize, b'\x00', 1)
-        self.assertRaises(ValueError, deserialize, [0, 256])
+        for a in bitarray('0111', 'big'), frozenbitarray('0111', 'big'):
+            self.assertEqual(serialize(a), b'\x14\x70')
 
-        b = b'\x03\x06'
-        for s in b, bytearray(b), memoryview(b), iter(b):
+    def test_deserialize_args(self):
+        for x in 0, 1, False, True, None, u'', u'01', 0.0, [0, 1]:
+            self.assertRaises(TypeError, deserialize, x)
+        # no arguments
+        self.assertRaises(TypeError, deserialize)
+        # too many arguments
+        self.assertRaises(TypeError, deserialize, b'\x00', 1)
+
+        blob = b'\x03\x06'
+        x = bitarray()  # we can deserialize a bitarray as it has a buffer
+        x.frombytes(blob)
+        for s in blob, bytearray(blob), memoryview(blob), x:
             a = deserialize(s)
+            self.assertEqual(a.to01(), '01100')
             self.assertEqual(a.endian(), 'little')
-            self.assertEqual(a, bitarray('01100'))
 
     def test_invalid_bytes(self):
         self.assertRaises(ValueError, deserialize, b'')
@@ -1480,21 +1925,25 @@ class TestsSerialization(unittest.TestCase, Util):
 
     def test_bits_ignored(self):
         # the unused padding bits (with the last bytes) are ignored
-        for b, a in [
-                (b'\x07\x01', bitarray('1', 'little')),
-                (b'\x07\x03', bitarray('1', 'little')),
-                (b'\x07\xff', bitarray('1', 'little')),
-                (b'\x17\x80', bitarray('1', 'big')),
-                (b'\x17\xc0', bitarray('1', 'big')),
-                (b'\x17\xff', bitarray('1', 'big')),
+        for blob, endian in [
+                (b'\x07\x01', 'little'),
+                (b'\x07\x03', 'little'),
+                (b'\x07\xff', 'little'),
+                (b'\x17\x80', 'big'),
+                (b'\x17\xc0', 'big'),
+                (b'\x17\xff', 'big'),
         ]:
-            self.assertEQUAL(deserialize(b), a)
+            a = deserialize(blob)
+            self.assertEqual(a.to01(), '1')
+            self.assertEqual(a.endian(), endian)
 
     def test_random(self):
         for a in self.randombitarrays():
-            b = deserialize(serialize(a))
-            self.assertEQUAL(a, b)
-            self.check_obj(b)
+            b = serialize(a)
+            c = deserialize(b)
+            self.assertEqual(a, c)
+            self.assertEqual(a.endian(), c.endian())
+            self.check_obj(c)
 
 tests.append(TestsSerialization)
 
